/**
 * The Paper Tape Project -- Punching subsystem
 * CLI Frontend implementation
 *
 * This is a very simple cross-platform stand alone command
 * line interface, with basic features (no [n]curses, etc.),
 * like the old "userspace-driver" had them. This includes:
 *
 *  * Reading bytes from stdin and transfering to the backend
 *  * Accepting some command line arguments for setting the file
 *    size (how much bytes we can await at stdin) or switching
 *    on DEBUG mode.
 *  * Giving a very simple progress output: For every byte
 *    one line.
 *
 * (c) 2008 Sven Kppel
 *
 **/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> /* Problematisch auf Windows! */
#include <errno.h>
#include <stdarg.h> /* var argument list */
#include <math.h>

#include "backend.h"

#define DPRINTF(bla...) { if(debug_flag) fprintf(stderr,bla); }
int debug_flag; /* global var. */

void puncher_frontend_visualize_byte(unsigned char buf);
void puncher_frontend_calculate_time(int already_punched_bytes, int estimated_punch_length, int mtime_since_beginning);

// We need a portable function for this one!
int puncher_frontend_mtime(long long since) {
        /* gibt millisekunden zurueck seit since. since sind dabei die
           Millisekunden seit der Unix-Epoche. Ein initiales since kann
           per mtime(0) gefunden werden. */
        struct timeval time;
        gettimeofday(&time, NULL);
              // Sec*10^6
        return (time.tv_sec * 1000000 + time.tv_usec) - since;
}

int main(int argc, char **argv) {
	long long time_since_beginning; // measuring time.
	int already_punched = 0; // Anzahl bereits gepunchte bytes
	int estimated_data_length = 0; // per Schalter -l angegebbare Datenlaenge
	size_t ret; // multi purpose return value
	
	printf("The Paper Tape Project -- Punching subsystem\n");
	printf("Cross Plattform Command line interface\n");
	printf("time ms |123.45678|  hex=dec status: =) punch successful / :-( still busy\n");
	
	// Starting of parsing program arguments
	debug_flag = 0;
        opterr = 0;
        int c;
        while( (c = getopt(argc, argv, "dhl:s:")) != -1)
            switch(c) {
                case 'd':
                    debug_flag = 1;
                    break;
                case 'h':
                    fprintf(stderr, "Usage: %s [-d] [-h] [-lNUM]\n"
                      "  optional parameters:\n"
                      "   -d   debug mode, print out verbose informations about signals\n"
                      "        if not set, only normal verbosity will be printed out on\n"
                      "        stdout, painting an ASCII paper tape which visualizes the\n"
                      "        bytes (see above for the caption)\n\n"
                      "   -h   display this help message\n\n"
                      "   -s   same as -l\n"
                      "   -l   sets the NUMBER OF BYTES which are expected on stdin.\n"
                      "        with this information we can complete the status information\n"
                      "        with useful informations about the estimated remaining time\n"
                      "        to punch (useful for longer data)\n"
                      "        example: cat 2k-data-file | %s -l2000\n",
                      argv[0], argv[0]);
                    return 0;
                case 'l':
                case 's':
                    c = atoi(optarg);
                    if( c > 0 ) {
                        estimated_data_length = c;
                        break;
                    }
                    printf("%s: data length must be a positive integer, like '123', not %s\n", argv[0], optarg);
                case '?': break;
            }
        // end of parsing program arguments

        // initializing backend...
        PuncherBackend* puncher;
        if(! (puncher = puncher_backend_new(debug_flag)) ) {
		perror("creating puncher backend");
        }

	/* The main loop (reading stdin and printing out) */
	/* We will just read 1 byte at one time to handle
	 * it nicely... */
	DPRINTF("Start reading from stdin...\n");
	time_since_beginning = puncher_frontend_mtime(0);
	for(;;) {
		unsigned char buf; /* stdin 1 byte buffer */

		ret = read(0, &buf, 1); /* 1 byte from stdin */
		if(ret < 0) {
			perror("read stdin");
			puncher_backend_destroy(puncher);
			return 1;
		}

		if(ret == 0) {/* EOF, end of input, etc. */
			DPRINTF("stdin EOF\n");
			break;
		}

		/* Visualisation of the byte in a stdout line */
		printf("%5ims ", (int)(puncher_frontend_mtime(time_since_beginning)/1000));
		puncher_frontend_visualize_byte(buf);
		printf(" 0x%02x=%03i ", buf, buf);
		puncher_frontend_calculate_time(++already_punched, estimated_data_length, (int)(puncher_frontend_mtime(time_since_beginning)/1000));

                /* call the backend */
                ret = puncher_backend_write_byte(puncher, buf);
		if(ret < 0) {
			perror("calling write backend");
			puncher_backend_destroy(puncher);
			return 1;
		} else if(ret == 0) {
		        printf("=)\n"); // everything okay
		} else if(ret == 1) {
		        printf(":-(\n"); // punch not okay, but go on.
		}
	} /* for stdin */

	DPRINTF("Transmission finished\n");
	if(puncher_backend_destroy(puncher)) {
		perror("finishing backend");
		return 1;
	} else
		return 0;
}

void puncher_frontend_visualize_byte(unsigned char buf) {
	/* Draw a papertape byte (like the perl program) */
	unsigned char against = 0x01;
	unsigned char check;
	int pos;
	
	printf("|");
	for(pos=0; pos < 8; pos++) {
		if(pos == 3) /* Streifenfuehrung */
			printf(".");
		check = buf;
		check >>= pos;
		
		if((check & against) == 0)
			printf(" "); /* bit nicht gesetzt */
		else
			printf("*"); /* bit gesetzt */
	} /*for */
	printf("|");
}


void puncher_frontend_calculate_time(int already_punched_bytes, int estimated_punch_length, int mtime_since_beginning) {
	/* In jeder Zeile berechnen, wie lange noch gepuncht wird,
	 * und ausgeben. Nur wenn Laenge per "-l" oder "-s" angegeben
	 * wird. In Bytes.
	 * 
	 * Irgendwie ist der Berechnungsalgorithmus etwas schrottig und
	 * produziert in den ersten Sekunden nur Mll, aber ansonsten
	 * ist er ganz okay.
	 */
	int percentage, remaining_sec = 0;
	 
	if(!estimated_punch_length) return; // wenn es gar nicht abgegeben wurde
	
	//printf("%i %i ", already_punched_bytes, estimated_punch_length, mtime_since_beginning);
	
	percentage = (int)rint(((float)already_punched_bytes / (float)estimated_punch_length) * 100);
	if(mtime_since_beginning != 0) // erst nach dem ersten Zyklus berechenbar
	    remaining_sec = (int)rint(((float)mtime_since_beginning / ((float)already_punched_bytes / (float)estimated_punch_length) - mtime_since_beginning)/1000);
	
	if(already_punched_bytes <= estimated_punch_length) {
	       printf("| %i%% (%i sec remaining)  ", percentage, remaining_sec);
	}
}
