/**
 * The GTK frontend to the puncher backends.
 * 
 * It uses GtkPaperTape to display the current open file and
 * serves several controls to control the punch process.
 *
 * started at 14.06.2008
 * (c) Sven
 *
 *
 *
 **/


#include <gtk/gtk.h>
#include <math.h>
#include <stdlib.h>
#include "backend.h"
#include "../visualisator/gtkpapertape.h"

#define _(a) a

/* while punching in the gtk main loop, update the gui! */
#define UPDATE_GTK_GUI  { while (gtk_events_pending ()) gtk_main_iteration (); }

/* The master structure... */
struct puncher_frontend {
	/* The state of our punching process */
	enum PUNCHER_FRONTEND_PROGRESS_STATE {
		PROGRESS_RUNNING,
		PROGRESS_PAUSED,
		PROGRESS_STOPPED
	} state;
	
	/* Various GtkWidgets: */
	GtkWidget *progress_bar;		// The GtkProgressBar
	
	// Label Grid:
	GtkWidget *progress_holes_total;	// counts ALL available bytes
	GtkWidget *progress_holes_punched;	// counts the already punched bytes
	GtkWidget *progress_holes_todo;		// counts the bytes which must be punched
	
	GtkWidget *progress_total_time;		// displays the calculated total punch time
	GtkWidget *progress_estimated_time;	// displays the estimated time until finish
	GtkWidget *progress_elapsed_time;	// displays the elapsed time since beginning
	
	GtkWidget *progress_time_per_hole;	// displays the time needed for one hole
	GtkWidget *progress_frequency;		// displays the amount of holes per second

	// Master buttons
	GtkWidget *button_start;		// The button to start punching
	GtkWidget *button_pause;		// The button to pause punching
	GtkWidget *button_stop;			// The button to stop punching

	// The papertape context
	GtkPaperTape *papertape;

	// This is supposed to be a link to the data
	// saved somewhere in GtkPaperTape/LOCHSTREIFEN*.
	// Don't even think about to free() it. The memory
	// leak doesn't hurt ;-)
	int length;
	byte_t *data;

	int punched;  // the already punched holes
	GTimer *timer; // the timer which measures
} d;

gboolean progress_update_time();

void puncher_frontend_prepare(GtkWidget *window) {
	/**
	 * This function will prepare the big struct puncher_frontend which is
	 * globally known as "d"
	 *
	 **/

	d.state = PROGRESS_STOPPED;
	d.length = 0;
	d.data = 0;
	d.punched = 0;
	
	d.papertape = GTK_PAPER_TAPE(gtk_paper_tape_new(window));
	
	d.progress_bar = gtk_progress_bar_new();
	
	#define PUNCHER_FRONTEND_FAST_LABEL(a) { d.a = gtk_label_new("x"); \
		gtk_label_set_selectable(GTK_LABEL(d.a), TRUE); }
	PUNCHER_FRONTEND_FAST_LABEL(progress_holes_total);
	PUNCHER_FRONTEND_FAST_LABEL(progress_holes_punched);
	PUNCHER_FRONTEND_FAST_LABEL(progress_holes_todo);
	PUNCHER_FRONTEND_FAST_LABEL(progress_total_time);
	PUNCHER_FRONTEND_FAST_LABEL(progress_estimated_time);
	PUNCHER_FRONTEND_FAST_LABEL(progress_elapsed_time);
	PUNCHER_FRONTEND_FAST_LABEL(progress_time_per_hole);
	PUNCHER_FRONTEND_FAST_LABEL(progress_frequency);

	d.button_start = gtk_button_new_with_mnemonic(_("Punch!")); 
	gtk_button_set_image(GTK_BUTTON(d.button_start),
		gtk_image_new_from_stock(GTK_STOCK_MEDIA_PLAY, GTK_ICON_SIZE_BUTTON));

	d.button_pause = gtk_button_new_with_mnemonic(_("Pause")); 
	gtk_button_set_image(GTK_BUTTON(d.button_pause),
		gtk_image_new_from_stock(GTK_STOCK_MEDIA_PAUSE, GTK_ICON_SIZE_BUTTON));

	d.button_stop = gtk_button_new_with_mnemonic(_("Stop")); 
	gtk_button_set_image(GTK_BUTTON(d.button_stop),
		gtk_image_new_from_stock(GTK_STOCK_MEDIA_STOP, GTK_ICON_SIZE_BUTTON));

	d.timer = g_timer_new();
}

void puncher_frontend_progress_set() {
	/**
	 * set information about holes
	 *
	 **/
	gtk_label_set_markup(GTK_LABEL(d.progress_holes_total),
		g_markup_printf_escaped("Loecher: <b>%3d</b>", d.length)
	);
	gtk_label_set_markup(GTK_LABEL(d.progress_holes_punched),
		g_markup_printf_escaped("Gestanzt: <b>%3d</b>", d.punched)
	);
	gtk_label_set_markup(GTK_LABEL(d.progress_holes_todo),
		g_markup_printf_escaped("Zu Stanzen: <b>%4d</b>", d.length - d.punched)
	);
}

gboolean puncher_frontend_puncher_start(GtkWidget *start_widget, gpointer nada) {
	/**
	 * The callback for the "start punching" button. This will prepare
	 * the gui, prepare observing and progress functions and of course
	 * this will start the low level punching driver.
	 **/
	
	// set the values
	d.state = PROGRESS_RUNNING;
	
	// prepare the gui
	gtk_widget_set_sensitive(d.button_start, FALSE);
	gtk_widget_set_sensitive(d.button_pause, TRUE);
	gtk_widget_set_sensitive(d.button_stop, TRUE);
	gtk_widget_set_sensitive(d.progress_bar, TRUE);
	
	// start the progress indicators:
	//g_timer_start(d.timer);
	//g_timeout_add(1000, progress_update_time, NULL);
	// kein Sinn - regelmaessig nach jedem punch machen!
	
	// start the punching loop
	puncher_frontend_punch_loop();
}

gboolean puncher_frontend_punch_loop() {
	/**
	 * The Punch main loop, as the gui drives it.
	 * Returns success or not (FALSE).
	 **/
	
	int x;
	
	for(x = 0; x < d.length; x++) {
		puncher_frontend_progress_set();
		puncher_frontend_progress_update_time();
		gtk_paper_tape_set_highlight(d.papertape, x, TRUE); // TRUE: Abhngig von Checkbox!
		UPDATE_GTK_GUI;
		punch_byte(d.data[x]);
		d.punched++;
	}
	
	gtk_paper_tape_remove_highlight(d.papertape);
	return TRUE;
}

gboolean puncher_frontend_progress_update_time() {
	/**
	 * This does exactly what the low level function "calculate_time" does;
	 * using the power of glib (GTimer). Furthermore additional values are
	 * calculated.
	 * 
	 * expects the GTimer to be started.
	 **/

	int percentage = (int)rint(((float)d.punched / (float)d.length) * 100);
	int sec_since_beginning = (int)rint(g_timer_elapsed(d.timer, NULL));
	int sec_total;
	int est = 0;
	
	gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(d.progress_bar),
		(double)d.punched / (double)d.length);
	gtk_progress_bar_set_text(GTK_PROGRESS_BAR(d.progress_bar),
		g_strdup_printf("%d%%", percentage)
	);
	
	// calc after first cycle
	if(sec_since_beginning != 0) {
		est = (int)rint((float)sec_since_beginning / ((float)d.punched / (float)d.length));
		
		gtk_label_set_markup(GTK_LABEL(d.progress_estimated_time),
			g_markup_printf_escaped("<b>%d</b> sec", est)
		);
	}
	
	gtk_label_set_markup(GTK_LABEL(d.progress_elapsed_time),
		g_markup_printf_escaped("%d", sec_since_beginning)
	);
	
	return !(est == sec_since_beginning && percentage == 100);
	// true zurueckgeben, wenns weitergehen soll.
}


gboolean puncher_frontend_open_file_dialog(GtkWidget *menuitem, GtkPaperTape* papertape) {
	/**
	 * displays an "open file" dialog which runs gtk_paper_tape_read_from_file() if success,
	 * is run by the menu item "open file"
	 **/

	GtkWidget *chooser;
	chooser = gtk_file_chooser_dialog_new(
		_("Select (binary) file to display as a paper tape"),
		GTK_WINDOW(papertape->parent_window),
		GTK_FILE_CHOOSER_ACTION_OPEN,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_OK,
		NULL);

	if(gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_OK) {
		char *filename;
		FILE *file;

		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (chooser));
		if(filename == NULL) {
			g_printf(_("No file selected to open.\n"));
			gtk_widget_destroy(GTK_WIDGET(chooser));
			return FALSE;
		} else {
			gtk_paper_tape_read_from_file(papertape, filename);
			
			d.length = papertape->lochstreifen->data_length;
			d.data = papertape->lochstreifen->data;
			
			g_free(filename);
			gtk_widget_destroy(GTK_WIDGET(chooser));
			return TRUE;
		}
	}
	gtk_widget_destroy(chooser);
	return FALSE;
} // open_file_dialog

void puncher_frontend_quit_application(GtkWidget *window, gpointer *nada) {
	puncher_backend_finish();
	exit(0);
	// on window close: manage punching interaction!!
}

int main(int argc, char *argv[]) {
	GtkWidget *window, *widget, *main_box, *tape_box, *punch_box, *statusbar;
	gtk_init(&argc, &argv);
	
	// main window generation
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_default_size(GTK_WINDOW(window), 600, 600);
	gtk_window_set_title(GTK_WINDOW(window), _("Papertape Puncher"));
	g_signal_connect_swapped(G_OBJECT(window), "destroy", G_CALLBACK(puncher_frontend_quit_application), NULL);
	
	// above-main-box generation (contains menubar + mainbox)
	puncher_frontend_prepare(window);
	widget = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(window), widget);
	
	// menus generation start
	{
		GtkWidget *menu, *menubar, *menuitem;
		// menus
		menubar = gtk_menu_bar_new();
		gtk_box_pack_start(GTK_BOX(widget), menubar, FALSE, TRUE, 0);
		gtk_widget_show(menubar);
	
		// FIRST menu: File {{{
		menu = gtk_menu_new();
		// Oeffnen
		g_signal_connect(G_OBJECT(fast_stock_menuitem(menu, GTK_STOCK_OPEN)),
			"activate", G_CALLBACK(puncher_frontend_open_file_dialog), d.papertape);
		fast_menu_seperator(menu);
		gtk_paper_tape_menu_export(d.papertape, menu);
		fast_menu_seperator(menu);
		g_signal_connect(G_OBJECT(fast_stock_menuitem(menu, GTK_STOCK_QUIT)),
			"activate", G_CALLBACK(gtk_main_quit), NULL);
	
		menuitem = gtk_menu_item_new_with_mnemonic("_Datei");
		gtk_menu_item_set_submenu(GTK_MENU_ITEM (menuitem), menu);
		gtk_menu_shell_append(GTK_MENU_SHELL (menubar), menuitem);
		gtk_widget_show (menuitem);
	
		// Zweites Menue: Ansicht
		menu = gtk_menu_new();
		gtk_paper_tape_menu_view(d.papertape, menu);
	
		menuitem = gtk_menu_item_new_with_mnemonic("_Ansicht");
		gtk_menu_item_set_submenu(GTK_MENU_ITEM (menuitem), menu);
		gtk_menu_shell_append(GTK_MENU_SHELL (menubar), menuitem);
		gtk_widget_show (menuitem);
	
		// Drittes Menue: Farben
		menu = gtk_menu_new();
		gtk_paper_tape_menu_colors(d.papertape, menu);
	
		menuitem = gtk_menu_item_new_with_mnemonic("_Farben");
		gtk_menu_item_set_submenu(GTK_MENU_ITEM (menuitem), menu);
		gtk_menu_shell_append(GTK_MENU_SHELL (menubar), menuitem);
		gtk_widget_show (menuitem);
	}
	// menus end
	
	// main box generation
	main_box = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(main_box), 10);
	gtk_container_add(GTK_CONTAINER(widget), main_box);
	
	// tape box generation
	widget = gtk_frame_new("udo-14.06.2008-generated-file.bin");
	gtk_box_pack_start(GTK_BOX(main_box), widget, TRUE, TRUE, 0);
	tape_box = GTK_WIDGET(gtk_paper_tape_get_whole_box(d.papertape));
	gtk_container_set_border_width(GTK_CONTAINER(tape_box), 10);
	gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(d.papertape->statusbar), FALSE);
	gtk_container_add(GTK_CONTAINER(widget), tape_box);

	// punch box generation
	widget = gtk_frame_new(_("Punching controls"));
	gtk_box_pack_start(GTK_BOX(main_box), widget, FALSE, TRUE, 0);
	punch_box = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(punch_box), 10);
	gtk_container_add(GTK_CONTAINER(widget), punch_box);
	
	// fill punch box:
	{
		GtkWidget *label;
		GtkWidget *holes_box, *progress_box, *button_box;
		
		holes_box = gtk_hbox_new(FALSE, 0);
		gtk_container_add(GTK_CONTAINER(punch_box), holes_box);
		
		gtk_container_add(GTK_CONTAINER(holes_box), d.progress_holes_total);
		gtk_container_add(GTK_CONTAINER(holes_box), d.progress_holes_punched);
		gtk_container_add(GTK_CONTAINER(holes_box), d.progress_holes_todo);

		progress_box = gtk_hbox_new(FALSE, 0);
		gtk_container_add(GTK_CONTAINER(punch_box), progress_box);
		
		/*label = gtk_label_new("Fortschritt: ");
		gtk_container_add(GTK_CONTAINER(progress_box), label);*/
		
		gtk_box_pack_start(GTK_BOX(progress_box), d.progress_bar, TRUE, TRUE, 0);
		
		label = gtk_label_new("Zeit: ");
		gtk_container_add(GTK_CONTAINER(progress_box), label);
		gtk_container_add(GTK_CONTAINER(progress_box), d.progress_estimated_time);
		gtk_container_add(GTK_CONTAINER(progress_box), d.progress_elapsed_time);
		
		// Button boxes
		button_box = gtk_hbutton_box_new();
		gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box), GTK_BUTTONBOX_START);
		gtk_container_add(GTK_CONTAINER(punch_box), button_box);
		
		gtk_container_add(GTK_CONTAINER(button_box), d.button_start);
		g_signal_connect(G_OBJECT(d.button_start), "clicked", G_CALLBACK(gui_puncher_start), NULL);
		gtk_container_add(GTK_CONTAINER(button_box), d.button_pause);
		gtk_container_add(GTK_CONTAINER(button_box), d.button_stop);
	}

	// parse command options
	{
		char *filename = NULL;
		gboolean read_from_stdin = FALSE;
		GError *error = NULL;
		GOptionContext *context;
		byte_t show_mystart = 0; // eigene Startprozedur anzeigen?
		GOptionEntry option_entries[] = {
			{ "filename", 'f', 0, G_OPTION_ARG_FILENAME, &filename, _("file to open"), NULL },
			{ "stdin", 's', 0, G_OPTION_ARG_NONE, &read_from_stdin, _("read from standard input"), NULL },
			{ NULL }
		};

		context = g_option_context_new(_(" - GtkPaperTape visualisation"));
		g_option_context_add_main_entries(context, option_entries, NULL);
		g_option_context_add_group(context, gtk_get_option_group(TRUE));
		g_option_context_parse(context, &argc, &argv, &error);

		if(read_from_stdin) {
			printf(_("%s: Reading data from standard input, will open window after EOF.\n"), argv[0]);
			byte_t *data;
			int length = file_get_contents(stdin, &data);
			gtk_paper_tape_set_data(d.papertape, length, data);
		} else if(filename != NULL) {
			// eine Datei einlesen
			printf(_("Reading from file '%s'\n"), filename);
			if(!gtk_paper_tape_read_from_file(d.papertape, argv[1])) {
				printf(_("Failed to read from file!\n"));
				return 1;
			}
		}
	}
	// end of parsing command options

	gtk_widget_show_all(window);
	gtk_main();
	return 0;
}
